// Corpus.cpp: implementation of the CCorpus class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#define _CRTDBG_MAP_ALLOC
#include "Corpus.h"
#ifdef _WIN32
#include "../XairaVersion.h"
#endif
//#include <xercesc/dom/deprecated/DOM_NamedNodeMap.hpp>
#include "SaraIndexContentHandler.h"
#include "TextReader.h"
#include "Logger.h"
#include "MemoryErr.h"

#ifdef _DEBUG
#define mynew ::new(_NORMAL_BLOCK,__FILE__,__LINE__)
#define mydelete ::delete
#else
#define mynew new
#define mydelete delete
#endif

void FatalError(const char*,const char*);
void FatalError(const char* s,const XMLCh*);

#ifndef _WIN32
#define SEP '/'
#else
#define SEP '\\'
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CCorpus::CCorpus(bool bD,bool bF,bool bH, bool bM,int kLoc, int kStr,char* cl,char* sBrk):m_bDump(bD),m_bFreq(bF), 
		m_bHdr(bH), m_bMem(bM), m_kLoc(kLoc), m_kStr(kStr), m_cl(cl), m_sBrk(sBrk)
{
	m_nfiles=0;
	m_files=NULL;
	m_bDiskErr=m_bMemoryErr=m_bUnkErr=m_bCreateErr=m_bHdrErr=false;
	m_bHdr=true;
}

CCorpus::~CCorpus()
{
	if (m_files) mydelete [] m_files;
}

void CCorpus::IndexPhase1()
{ 
	CIndexer ind(this,m_kLoc,m_kStr);
	m_ind=&ind;
	ind.m_bMem=m_bMem;
	m_errorcount = 0; m_skip=0; m_nif=0;
	ind.SetBreak(m_sBrk);
	ind.m_fe << L("<log corpus=\"") << (const XMLCh*) Param(PARAM_NAME).getTerminatedBuffer() << L("\">") << chLF; 
	const char* loc="";
	_lang* la=m_r.m_deflang;
	if (la) {
		loc=la->locale;
		ind.SetLang(la->locale);
	}
	else ind.SetLang();
	Options(&ind);
	ind.Stats();
	try {
		if (!ReadHeader()) {
				printf("Fatal error, see log for details\n");
				ind.Log(L("fatal-error"),L("Header error, cannot continue"));
				ind.EndLog();
				m_bHdrErr=true;
				return;
		}
		ind.StartCorpus(Param(PARAM_NAME),m_nfiles,la);
		ind.Log(L("process"),L("Phase 1 (%d files)"),m_nfiles+2);
//		DefSCs();
		m_parser = XMLReaderFactory::createXMLReader();
		bool bValid=Param(PARAM_VALIDATE)=="on"?true:false; // sigh
		m_parser->setFeature(xercesc::XMLUni::fgSAX2CoreValidation, bValid);
		m_parser->setFeature(xercesc::XMLUni::fgXercesDynamic, true);
		if (Param(PARAM_GRAMMARCACHE)=="on") {
			m_parser->setFeature(XERCES_CPP_NAMESPACE::XMLUni::fgXercesCacheGrammarFromParse, true);
			m_parser->setFeature(XERCES_CPP_NAMESPACE::XMLUni::fgXercesUseCachedGrammarInParse, true);
		}
		SaraIndexContentHandler handler;
		handler.SetOutput(&ind);
        m_parser->setContentHandler(&handler);
        m_parser->setErrorHandler(&handler);
        m_parser->setLexicalHandler(&handler);
        m_parser->setDeclarationHandler(&handler);
		m_parser->setEntityResolver(&handler);
		m_parser->setExitOnFirstFatalError(false);
		int i;
		for (i=0;i<GetCount();i++) {
			handler.Reset();
			bool b=IndexFile(GetPath(i),GetRelPath(i),GetName(i),i,&ind);
//			ind.Log(L"stat",L"Max depth %d",ind.m_dict.GetDepth());
			if (i < 2 && !b) {
				printf("Fatal error, see log for details\n");
				ind.Log(L("fatal-error"),L("Header or bibliography missing"));
				ind.EndLog();
				return;
			}
		}
		delete m_parser;
		ind.EndCorpus();
		int k=ind.ErrCount();
		if (k > 0) ind.Log(L("process"),L("Errors in texts, see log"));
		ind.Log(L("process"),L("Writing grammar"));
		ind.DSC();
		ind.Phase2();
	}
	catch( std::bad_alloc &ba) {
		printf("Fatal error, see log for details\n");
		ind.Log(L("fatal-error"),L("Out of memory (%S)"),ba.what());
		ind.m_bMem=true;
		ind.Stats();
		m_bMemoryErr=true;
		ind.EndLog();
		return;
    }
	catch (CMemoryErr m)
	{
		printf("Fatal error, see log for details\n");
		ind.Log(L("fatal-error"),L("Out of memory (%S,%dx%d)"),m.m_alloc,m.m_k1,m.m_k2);
		ind.m_bMem=true;
		ind.Stats();
		m_bMemoryErr=true;
		ind.EndLog();
		return;
	}
	catch (CDiskWriteErr m)
	{
		printf("Fatal error, see log for details\n");
		ind.Log(L("fatal-error"),L("Disk write error (%d, %d, %S)"),m.m_k1,m.m_k2,m.m_alloc);
		m_bDiskErr=true;
		ind.EndLog();
		return;
	}
	catch (CDiskCreateErr* e)
	{
		ind.Log(L("fatal-error"),L("Error creating %s - %S"),e->m_path,e->m_err);
		ind.EndLog();
		m_bCreateErr=true;
		return;
	}
	catch (SAXException& e)
	{
		ind.Log(L("parse-error"),L("Parse error %s"),e.getMessage());
		m_bUnkErr=true;
		return;
	}
	catch (char* s)
	{
		printf("Fatal error, see log for details\n");
		ind.Log(L("fatal-error"),L("Unknown error in indexer (%S)"),s);
		ind.EndLog();
		m_bUnkErr=true;
		return;
	}
	catch (...)
	{
		printf("Fatal error, see log for details\n");
		ind.Log(L("fatal-error"),L("Unknown error in indexer(%S)"),strerror(errno));
		ind.EndLog();
		m_bUnkErr=true;
		return;
	}
	ind.EndLog();
}


void CCorpus::StartEntity(UnicodeString name,UnicodeString relpath)
{
	m_nest++;
	for (int i=0;i<m_nest;i++) {
		wprintf(L("   "));
		wprintf(L("%s\n"),name.getTerminatedBuffer());
	}
	CTextReader* r=mynew CTextReader(exppath(PARAM_TEXTPATH,relpath),m_ind);
	int nb;
	bool b=r->Read(nb);
	assert(b);
	m_nb+=nb;
	m_ind->Push();
	m_ind->m_reader=r;
	m_ind->m_relpath=relpath;
}

void CCorpus::EndEntity()
{
	m_ind->Pop();
	m_nest--;
}



bool CCorpus::IndexFile(UnicodeString path,UnicodeString relpath,UnicodeString name,int i,CIndexer* ind)
{
	m_nest=0;
	clock_t start = clock();
	ind->StartText(name,relpath);
//	wprintf(L"%s",name.getTerminatedBuffer());
	CTextReader r(path,ind);
	ind->m_reader=&r;
	int nb;
	if (!r.Read(nb)) {
//		wprintf(L" - cannot index\n");
		ind->SkipText(name);
		return false;
	}
	m_nb=nb;
	ind->NewFile(i);
	InputSource* is=r.GetInputSource();
    m_parser->parse(*is);
	mydelete is;
	m_errorcount += m_parser->getErrorCount()+ind->ErrCount();
	ind->EndText(name,m_parser->getErrorCount());
	clock_t finish = clock();
	double duration = (double)(finish - start) / CLOCKS_PER_SEC;

	m_ind->Log(L("stat"),L("<stat name=\"rate\" i=\"%d\">%d</stat>"),i,(int) (m_nb/duration));

	return true;
//	delete parser;
}

int CCorpus::Status()
{
	if (m_bHdrErr) return 5;
	if (m_bMemoryErr) return 8;
	if (m_bDiskErr) return 9;
	if (m_bUnkErr) return 10;
	if (m_bCreateErr) return 11;
	if (m_skip > 0) return 6;
	if (m_errorcount > 0) return 7;
	return 0;
}

int CCorpus::GetCount()
{
	return m_nfiles+2;
}

UnicodeString CCorpus::GetPath(int i)
{
	if (i==0) {
		return Param(PARAM_HEADER);
	} else if (i==1) {
		return Param(PARAM_BIB);
	}
	else return exppath(PARAM_TEXTPATH,m_files[i-2]);
}

UnicodeString  CCorpus::GetBibPath()
{
	return Param(PARAM_BIB);
}

UnicodeString CCorpus::GetRelPath(int i)
{
	if (i<2) {
		return (const UChar*) L("/");
	}
	else return m_files[i-2];
}

UnicodeString CCorpus::GetBibRelPath()
{
	return GetBibName();
}

UnicodeString CCorpus::GetName(int i)
{
	if (i==0) return (const UChar*) L("header");
	else if (i==1) return (const UChar*) L("bibliography");
	return m_files[i-2];
}

UnicodeString CCorpus::GetBibName()
{
	return param_vals[PARAM_BIB];
}

void CCorpus::TestIndexDir()
{
	xairo_uchar path[_MAX_PATH];
	wcscpy(path,(const xairo_uchar*) exppath(CCorpus::PARAM_INDEXPATH,L("xid.xml")).getTerminatedBuffer());
	FILE* f=_wfopen(path,L("wb"));
	if (f == NULL) throw(CDiskWriteErr("id",0,0));
	fprintf(f,"<xairo_index ver=\"%d\"/>",DBVER);
	fclose(f);
}

bool he(header_error,const xairo_uchar* m,void* v) {
	CCorpus* c=(CCorpus*) v;
	c->m_ind->HeaderError(m);
//	c->m_ind->IncErrs();
	return true;
}

bool CCorpus::ReadHeader()
{
//	printf("0");
	m_r.SetErrorHandler(he,this);
	int k=m_r.Read(GetPath(0)); 
//	m_bib=m_r.ReadBib(Param(PARAM_BIB)); 
	if (k) return false;
	if (m_r.OldStyle()) {
		m_ind->HeaderError(L("Header has no xairaSpecification, old format?"));
		return false;
	}
	try {
		m_r.GetTagUsage(); 
		m_r.GetLabels(); 
		m_r.GetAddKeys(); 
		m_r.GetIPSpec(); 
		m_r.GetTaxonomies(); 
		m_r.GetLangs(); 
		m_r.GetCodeBooks();
	}
	catch(CHeaderError*)
	{
		return false;
	}
//	if (m_r.ErrCount() > 0) return false;
	for (int i=0;i<m_r.m_nwtags;i++) {
		_wtag* w=&m_r.m_wtags[i];
		_tagusage* t=m_r.GetTagUsage(w->s);
		t->ip=EIP_MAKE(EIP_NO,EIP_INHERIT,EIP_INHERIT);
	}
	for (int i=0;i<m_r.m_naddkeys;i++) {
		_addkey* a=&m_r.m_addkeys[i];
		if (a->isLemma && !a->isPseudo) {
			for (int j=0;j<a->ngi;j++) {
				_attusage* at=m_r.GetAttUsage(a->gi[j],a->att);
				at->ip=IP_NONE;
			}
		}
	}
	return true;
}

void CCorpus::ReadFileList()
{
	m_tfl=NULL; m_nfiles=0;
	XercesDOMParser d;
	d.setErrorHandler(&m_eh);
	d.parse((const xairo_uchar*) exppath(PARAM_ETCPATH,(const UChar*) L("file_index.xml")).getTerminatedBuffer());
	DOMDocument* doc=d.getDocument();
//	DOMNodeList* n=doc->getElementsByTagName(L"file"); // very slow if list is long
	DOMElement* root=doc->getDocumentElement();
	for (DOMNode* m=root->getFirstChild();m;m=m->getNextSibling())
	{
		if (m->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* me=(DOMElement*) m;
		const XMLCh* tag=me->getTagName();
		if (wcscmp(tag,L("file")) != 0) continue;

		if (wcscmp(me->getAttribute(L("disabled")),L("yes"))==0) continue;
		DOMNode* m1=m->getFirstChild();
		if (m1==NULL) continue;
		if (m1->getNodeType() != DOMNode::TEXT_NODE) continue;
		const XMLCh* s=m1->getNodeValue();
		AddFile((UChar*) s);
	}
	BuildFileArray();
}




void CCorpus::DHandle::fatalError(const SAXParseException& e)
{
	throw e;
}

void CCorpus::DHandle::error(const SAXParseException& e)
{
	throw e;
}


void CCorpus::ReadParameters(const char* pfile)
{
	char path[_MAX_PATH];
	_fullpath(path,pfile,_MAX_PATH);
	UnicodeString ppfile(path);
	int k=ppfile.lastIndexOf((UChar) SEP);
	if (k==-1) return;
	param_vals[PARAM_ROOTPATH]=UnicodeString(ppfile,0,k+1);
	XercesDOMParser d;
	d.setErrorHandler(&m_eh);
	d.parse((const xairo_uchar*) ppfile.getTerminatedBuffer());
	DOMDocument* doc=d.getDocument();
	DOMNodeList* n=doc->getElementsByTagName(L("parameter"));
	for (unsigned i=0;i<n->getLength();i++) {
		DOMNode* m=n->item(i);
		const XMLCh *x,*y;
		DOMNamedNodeMap* a=m->getAttributes();
		DOMNode* na=a->getNamedItem(L("name"));
		if (na==NULL) continue;
		x=na->getNodeValue();
		DOMNode* mm=m->getFirstChild();
		if (mm==NULL) continue;
		if (mm->getNodeType()!=DOMNode::TEXT_NODE) continue;
		y=mm->getNodeValue();
		SetParam((const UChar*) x,(const UChar*) y);
	}
}


char* CCorpus::param_names[PARAM_MAX]=
{"etcpath","name","textpath","rootpath","indexpath","headerpath","temppath","log","sourcepath","bibpath","validate","grammarcache"};

void CCorpus::SetParam(UnicodeString name,UnicodeString value)
{
	int i;
	for (i=0;i<PARAM_MAX;i++)
	{
		if (name==param_names[i]) break;
	}
	if (i==PARAM_ROOTPATH) return;
	if (i>=PARAM_MAX) return;
	param_vals[i]=value;
}

#define SEPCHAR(c) (c=='/'||c=='\\')


UnicodeString CCorpus::Param(param p)
{
	UnicodeString s=param_vals[p];
	if ((s.length() > 1 && s[1]==':') || p==PARAM_NAME) return s;
	if (p==PARAM_VALIDATE||p==PARAM_GRAMMARCACHE) return s;
	UnicodeString ss=param_vals[PARAM_ROOTPATH];
	ss.append(s);
	return ss;
}

UnicodeString CCorpus::exppath(param pa,UnicodeString s)
{
	UnicodeString p=Param(pa);
	int k=p.length();
	if (k > 0 && !SEPCHAR(p[k-1])) 
		p+=(const UChar*) L("/");
	p+=s;
	return p;
}




void CCorpus::AddFile(UnicodeString s)
{
	_tfl** t=&m_tfl;	
	while (*t) t=&(*t)->next;
	*t=mynew _tfl;
	(*t)->name=s;
	(*t)->next=NULL;
	m_nfiles++;
}

void CCorpus::BuildFileArray()
{
	m_files=mynew UnicodeString [m_nfiles];
	_tfl* t=m_tfl;
	for (int i=0;i<m_nfiles;i++) {
		m_files[i]=t->name;
		_tfl* tt=t->next;
		mydelete t;
		t=tt;
	}
}

void CCorpus::Options(CIndexer* i)
{
#ifndef _WIN32
	i->Log(L("note"),L("Xaira-indexer version " VERSION " built " __DATE__),VERSION,__DATE__);
#else
	i->Log(L("note"),L("Xaira-indexer version %d.%d.%d.%d built %S"),XV1,XV2,XV3,XV4,__DATE__);
#endif
	i->Log(L("note"),L("Command line options: %S"),m_cl);
	i->Log(L("note"),L("Language set to: %S"),i->GetLang());

}

void CCorpus::OptionLabel(CIndexer* i,_lab* l)
{
	UnicodeString s;
	if (l->isSpec) {
		switch (l->e) {
		case SPE_FILE:
			s=(const UChar*) L("xara:file, auto");
			break;
		case SPE_LINE:
			s=(const UChar*) L("xara:line, auto");
			break;
		}
	} else {
		s=l->elt[0];
		s+=(const UChar*) L(", ");
		if (l->isAuto) {
			s+=(const UChar*) L("auto");
		} else 
			s+=l->att;
	}
	m_ind->Log(L("note"),L("Label option: %s"),s.getTerminatedBuffer());
}

bool CCorpus::IllegalInFile(XMLCh w)
{
	if (w < 31) return true;
	switch (w) {
		case '<': case '>': case ':': case '\"': case '/': case '\\': case '|':
			return true;
		default:
			return false;
	}
}

UnicodeString CCorpus::NormaliseSCName(UnicodeString s)
{
	UnicodeString t=s;
	for (int i=0;i<t.length();i++) {
		if (IllegalInFile(t[i])) t.setCharAt(i,'_');
	}
	return t;
}

UnicodeString CCorpus::GetSCPath(UnicodeString tax,UnicodeString ext)
{
	UChar p[_MAX_PATH];
	UnicodeString s=NormaliseSCName(tax);
	swprintf((xairo_uchar*) p,L("%s.%s"),s.getTerminatedBuffer(),ext.getTerminatedBuffer());
	return exppath(PARAM_ETCPATH,p);
}

void CCorpus::DefSCs()
{
	int n=m_r.m_ntax;
	FILE* fr=iopen(exppath(PARAM_ETCPATH,(const UChar*) L("screg")),(const UChar*) L("w"));
//	int m=0;
	for (int i=0;i<n;i++) {
		UnicodeString id=m_r.m_taxonomies[i].id;
		FILE* f=iopen(GetSCPath(id,(const UChar*) L("sc")),(const UChar*) L("w"));
		fwprintf(f,L("0\n"));
		fclose(f);
//		UChar p[_MAX_PATH];
//		swprintf(p,L"%s.scn",id.getTerminatedBuffer());
//		f=iopen(exppath(PARAM_ETCPATH,p),L"w");
		f=iopen(GetSCPath(id,(const UChar*) "scn"),(const UChar*) L("w"));
		for (int j=0;j<m_r.m_taxonomies[i].n;j++) {
			ClassNames(f,&m_r.m_taxonomies[i].l[j]);
//			UnicodeString n=m_r.m_taxonomies[i].l[j].desc;
//			fwprintf(f,L"%s\n",n.getTerminatedBuffer());
		}
		fwprintf(f,L("%s\n"),m_r.m_taxonomies[i].unclass.getTerminatedBuffer());
		fclose(f);
		fwprintf(fr,L("%s %s\n"),NormaliseSCName(id).getTerminatedBuffer(),m_r.m_taxonomies[i].desc.getTerminatedBuffer());
	}
	fclose(fr);
}

void CCorpus::ClassNames(FILE* f,_class* l)
{
	fwprintf(f,L("%s\n"),l->desc.getTerminatedBuffer());
	for (int i=0;i<l->n;i++) {
		ClassNames(f,&l->l[i]);
	}
}


int CCorpus::NText()
{
	int n=m_ind->m_nt;
	if (n>0) n--;
	return n;
}